{-
Red Black Trees

-- Jonathan Leivent
-- 9-Jan-2012

This work is licensed under the Creative Commons
Attribution-ShareAlike 3.0 Unported License. To view a copy of this
license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send
a letter to Creative Commons, 444 Castro Street, Suite 900, Mountain
View, California, 94041, USA.
-}

module RedBlack where
open import Relation.Binary
open import Relation.Binary.PropositionalEquality

module RedBlackTree
  (A       : Set) -- elements
  (K       : Set) -- keys
  (†_      : A → K) -- get the key of an element
  (_<K_    : K → K → Set) -- ordering
  (_<K?_   : ∀ (x y : K) → Tri (x <K y) (x ≡ y) (y <K x)) -- ordering is tri-decidable
  (<Ktrans : ∀ {x y z : K} → x <K y → y <K z → x <K z) -- ordering is transitive
  where

  open import Data.Nat
  open import Data.Sum
  open import Data.Product
  open import Data.Unit using (⊤; tt)

  private
    -- for comparing As without constantly referring to their keys:
    _≤A_ : (a b : A) → Set
    a ≤A b = († a <K † b) ⊎ († a ≡ † b)

    _≤A?_ : (a b : A) → (a ≤A b) ⊎ (b ≤A a)
    a ≤A? b with † a <K? † b
    a ≤A? b | tri< lt _ _ = inj₁ (inj₁ lt)
    a ≤A? b | tri≈ _ eq _ = inj₁ (inj₂ eq)
    a ≤A? b | tri> _ _ gt = inj₂ (inj₁ gt)

    ≤Atrans : {a b c : A} → a ≤A b → b ≤A c → a ≤A c
    ≤Atrans (inj₁ x) (inj₁ x') = inj₁ (<Ktrans x x')
    ≤Atrans (inj₁ x) (inj₂ y) rewrite y = inj₁ x
    ≤Atrans (inj₂ y) bc rewrite y = bc

  -- The RedBlackInternal module is parameterized this rather messy
  -- way so that it can be instantiated both by an actual list
  -- interpretation of F with the sorted predicate based on ≤A in the
  -- usual way, as well as an "erased" interpretation where F and
  -- sorted F are both ⊤.  I did this roll-your-own erasing because I
  -- couldn't figure out how to get Agda's irrelevance mechanism to
  -- work.
  module RedBlackInternal
    (F        : Set) -- a monoid that acts like List A (think: "Flattening" of a tree)
    (↑_       : A → F) -- lift A into F, like "[ a ]"
    (_++_     : F → F → F) -- the F monoid's operator (list append)
    ([]       : F) -- the F monoid's identity
    ([]++     : (f : F) → [] ++ f ≡ f)
    (++[]     : (f : F) → f ++ [] ≡ f)
    (++assoc  : (a b c : F) → (a ++ b) ++ c ≡  a ++ (b ++ c)) -- associativity of F monoid operator
    (sorted   : F → Set) -- predicate that says whether an F is sorted under ≤A or not
    (sorted[] : sorted [])
    (sorted1  : (a : A) → sorted (↑ a))
    -- common required lemmas on sorted Fs [note: lemma-like functions are prefixed with £]
    (£sort₁   : (f₁ f₂ : F) → sorted (f₁ ++ f₂) → sorted f₁)
    (£sort₂   : (f₁ f₂ : F) → sorted (f₁ ++ f₂) → sorted f₂)
    -- lemmas about sorted Fs required for ins:
    (£ins₁    : (f₁ f₂ f₃ : F){x y : A} → sorted ((f₁ ++         f₂)  ++ (↑ y ++ f₃))
                                        → sorted  (f₁ ++ (↑ x ++ f₂))
                              → x ≤A y  → sorted ((f₁ ++ (↑ x ++ f₂)) ++ (↑ y ++ f₃)))
    (£ins₂    : (f₁ f₂ f₃ : F){y x : A} → sorted (f₁ ++ (↑ y ++ (f₂ ++         f₃)))
                                        → sorted                (f₂ ++ (↑ x ++ f₃))
                              → y ≤A x  → sorted (f₁ ++ (↑ y ++ (f₂ ++ (↑ x ++ f₃)))))
    --- ∉ is required for fnd and del:
    (_∉_      : (k : K)(f : F) → Set) -- predicate that says no element with key k is in f
    (£∉[]     : (k : K) → k ∉ []) -- nothing is in the identity
    -- lemmas about ∉ required for fnd and del:
    (£∉<      : {k : K}{x : A}(f : F) → k <K † x → sorted (↑ x ++ f) → k ∉ (↑ x ++ f))
    (£∉>      : {k : K}{x : A}(f : F) → † x <K k → sorted (f ++ ↑ x) → k ∉ (f ++ ↑ x))
    (£∉++     : (k : K){f₁ f₂ : F} → k ∉ f₁ → k ∉ f₂ → k ∉ (f₁ ++ f₂))
    -- lemmas about sorted Fs required for del:
    (£del₁    : (f₁ f₂ f₃ f₄ : F) → sorted ((f₁ ++ (f₂ ++ f₃)) ++ f₄)
                                  → sorted ((f₁ ++        f₃)  ++ f₄))
    (£del₂    : (f₁ f₂ f₃ : F) → sorted (f₁ ++ (f₂ ++ f₃)) 
                               → sorted (f₁        ++ f₃))
    (£del₃    : (f₁ f₂ f₃ f₄ f₅ : F) → sorted (f₁ ++ (f₂ ++ (f₃ ++ (f₄ ++ f₅))))
                                     → sorted (f₁ ++ (f₂ ++ (f₃ ++        f₅))))
    where

    -- an alias of _++_, because I couldn't figure out how to give an
    -- infix decl for module param _++_
    infixr 5 _⊞_
    _⊞_ : F → F → F
    _⊞_ = _++_

    data C : Set where -- an easy to remember encoding for colors
      ♥ : C -- hearts are red
      ♠ : C -- spades are black

    -- Red Black Trees
    --
    --   F is the "flattening" of the tree - an ordered list of all of
    --   the tree's elements (hopefully erasable).  Each tree node
    --   requires that its flattening is sorted, and is the
    --   concatenation of the flattenings of its subtrees, with its
    --   own element placed in between them.
    --
    --   C and ℕ are the standard red-black tree color and black level
    --   of the nodes, respectively.
    --
    --   Leaf nodes are black, have zero black level, and do not carry
    --   any data elements.
    --
    --   Red nodes require both subtrees to be black, black nodes do
    --   not place any color requirement on subtrees.
    data RB : F → C → ℕ → Set where
      leaf  :                                               RB [] ♠ zero
      ♥node : {f₁ : F}        {n : ℕ} → (t₁ : RB f₁ ♠ n)  →
                                        (d : A)           →
              {f₂ : F} →                (t₂ : RB f₂ ♠ n)  →
            let f = f₁ ⊞ ↑ d ⊞ f₂ in    (s : sorted f)    → RB f ♥ n
      ♠node : {f₁ : F}{c₁ : C}{n : ℕ} → (t₁ : RB f₁ c₁ n) →
                                        (d : A)           →
              {f₂ : F}{c₂ : C}        → (t₂ : RB f₂ c₂ n) →
            let f = f₁ ⊞ ↑ d ⊞ f₂ in    (s : sorted f)    → RB f ♠ (suc n)

    -- rewrite the flattening (F part) of a node
    infixl 20 _▷_
    _▷_ : {f₁ f₂ : F}{c : C}{n : ℕ} → RB f₁ c n → f₁ ≡ f₂ → RB f₂ c n
    t ▷ refl = t

    -- same as above, only for nodes in Σ's
    infixl 20 _Σ▷_
    _Σ▷_ : {f₁ f₂ : F}{n : ℕ} → Σ[ c ∶ C ](RB f₁ c n) → f₁ ≡ f₂ → Σ[ c ∶ C ](RB f₂ c n)
    t Σ▷ refl = t

    -- rewrite a sorted flattening
    infixl 20 _s▷_
    _s▷_ : {f₁ f₂ : F} → sorted f₁ → f₁ ≡ f₂ → sorted f₂
    s s▷ refl = s

    -- the sorted flattening of a node
    sortedOf : {f : F}{c : C}{n : ℕ} → RB f c n → sorted f
    sortedOf leaf = sorted[]
    sortedOf (♥node _ _ _ s) = s
    sortedOf (♠node _ _ _ s) = s

    -- conversion of a red node to a black node
    ♥⇛♠ : {f : F}{n : ℕ} → RB f ♥ n → RB f ♠ (suc n)
    ♥⇛♠ (♥node t₁ d t₂ y) = ♠node t₁ d t₂ y

    -- some more associativity lemmas:
    abstract
      £⊞₁ : (a b c d e f : F) → a ⊞ b ⊞ (c ⊞ d ⊞ e) ⊞ f ≡ (a ⊞ b ⊞ c) ⊞ d ⊞ e ⊞ f
      £⊞₁ a b c d e f rewrite ++assoc a (b ⊞ c) (d ⊞ e ⊞ f) | ++assoc b c (d ⊞ e ⊞ f) | sym (++assoc d e f) | sym (++assoc c (d ⊞ e) f) = refl

      £⊞₂ : (a b c d : F) → (a ⊞ b ⊞ c) ⊞ d ≡ a ⊞ b ⊞ c ⊞ d
      £⊞₂ a b c d rewrite ++assoc a (b ⊞ c) d | ++assoc b c d = refl

      £⊞₃ : (a b c d : F) → a ⊞ (b ⊞ c) ⊞ d ≡ a ⊞ b ⊞ c ⊞ d
      £⊞₃ a b c d rewrite ++assoc b c d = refl

    abstract
      £sort₂' : (f₁ : F)(x : A)(f₂ : F) → sorted (f₁ ⊞ ↑ x ⊞ f₂) → sorted f₂
      £sort₂' f₁ x f₂ s = £sort₂ (↑ x) f₂ (£sort₂ f₁ (↑ x ⊞ f₂) s)

    -- some re-balancing functions (more later, as needed)
    bal₀ : {f₁ : F}{c₁ : C}{n : ℕ} →       RB f₁ c₁ n →
                                           (x : A)    →
           {f₂ : F} →                      RB f₂ ♥ n  →
                                           (z : A)    →
           {f₃ : F}{c₃ : C} →              RB f₃ c₃ n →
         let f = f₁ ⊞ ↑ x ⊞ f₂ ⊞ ↑ z ⊞ f₃ in sorted f   → RB f ♥ (suc n)
    bal₀ {f₁}t₁ x (♥node {f₂₁}t₂₁ y {f₂₂}t₂₂ s₂) z {f₃}t₃ s rewrite £⊞₁ f₁ (↑ x) f₂₁ (↑ y) f₂₂ (↑ z ⊞ f₃)
      = ♥node (♠node t₁ x t₂₁ (£sort₁ (f₁ ⊞ ↑ x ⊞ f₂₁) (↑ y ⊞ (f₂₂ ⊞ ↑ z ⊞ f₃)) s)) 
            y (♠node t₂₂ z t₃ (£sort₂' (f₁ ⊞ ↑ x ⊞ f₂₁) y (f₂₂ ⊞ ↑ z ⊞ f₃) s)) s

    bal₁ : {f : F}{n : ℕ} →      RB f ♠ (suc n) →
                                 (D : A)        →
           {f₃ : F}{c₃ : C} →    RB f₃ c₃ n     →
         let f' = f ⊞ ↑ D ⊞ f₃ in sorted f'      → Σ[ c' ∶ C ](RB f' c' (suc n))
    bal₁ (♠node {f₁}{♥}t₁ y {f₂}t₂ s') D {f₃} t₃ s rewrite £⊞₂ f₁ (↑ y) f₂ (↑ D ⊞ f₃)
      = , ♥node (♥⇛♠ t₁) y (♠node t₂ D t₃ (£sort₂' f₁ y (f₂ ⊞ ↑ D ⊞ f₃) s)) s
    bal₁ (♠node {f₁}{♠}t₁ y {f₂}{♥}t₂ s') D {f₃} t₃ s rewrite £⊞₂ f₁ (↑ y) f₂ (↑ D ⊞ f₃)
      = , bal₀ t₁ y t₂ D t₃ s
    bal₁ (♠node {f₁}{♠}t₁ y {f₂}{♠}t₂ s') D t₃ s
      = , ♠node (♥node t₁ y t₂ s') D t₃ s

    bal₂ : {f₀ : F}{c₀ : C}{n : ℕ} → RB f₀ c₀ n     →
                                     (D : A)        →
           {f : F} →                 RB f ♠ (suc n) →
         let f' = f₀ ⊞ ↑ D ⊞ f in     sorted f'      →  Σ[ c' ∶ C ](RB f' c' (suc n))
    bal₂ {f₀}t₀ D (♠node {f₁}t₁ y {f₂}{♥}t₂ s') s rewrite sym (£⊞₂ f₀ (↑ D) f₁ (↑ y ⊞ f₂))
      = , ♥node (♠node t₀ D t₁ (£sort₁ (f₀ ⊞ ↑ D ⊞ f₁) (↑ y ⊞ f₂) s)) y (♥⇛♠ t₂) s 
    bal₂ t₀ D (♠node {f₁}{♥}t₁ y {f₂}{♠}t₂ s') s
      = , bal₀ t₀ D t₁ y t₂ s
    bal₂ t₀ D (♠node {f₁}{♠}t₁ y {f₂}{♠}t₂ s') s 
      = , ♠node t₀ D (♥node t₁ y t₂ s') s

    -- a comprehensive return type for ins
    data InsR : (x : A)(f : F)(c : C)(n : ℕ) → Set where
      ♥case :               (f₁' : F)
                            (f₂' : F)
              {f : F}       (e : f ≡ f₁' ⊞ f₂')
              {x : A}{n : ℕ}(t : RB (f₁' ⊞ ↑ x ⊞ f₂') ♠ (suc n))        → InsR x f ♥ n
      ♠case :               (f₁' : F)
                            (f₂' : F)
              {f : F}       (e : f ≡ f₁' ⊞ f₂')
              {x : A}{n : ℕ}(ct : Σ[ c ∶ C ](RB (f₁' ⊞ ↑ x ⊞ f₂') c n)) → InsR x f ♠ n

    -- Insert element x
    ins :                         (x : A)
          {f : F}{c : C}{n : ℕ} → RB f c n → InsR x f c n
    ins x leaf                                                             = ♠case [] [] (sym ([]++ [])) 
                                                                             (, ♥node leaf x leaf (sorted1 x s▷ sym (++[] (↑ x))
                                                                                                   s▷ sym ([]++ (↑ x ⊞ []))))
    ins x (♥node t₁ y t₂ s)  with x ≤A? y
    ins x (♥node t₁ y t₂ s)     | inj₁ x≤y with ins x t₁ 
    ins x (♥node t₁ y {f₂}t₂ s) | inj₁ x≤y    | ♠case f₁' f₂' refl (c , t) = ♥case f₁' (f₂' ⊞ ↑ y ⊞ f₂) (++assoc f₁' f₂' (↑ y ⊞ f₂))
                                                                             (♠node t y t₂ (£ins₁ f₁' f₂' f₂ s (sortedOf t) x≤y)
                                                                              ▷ ++assoc f₁' (↑ x ⊞ f₂') (↑ y ⊞ f₂) ▷ £⊞₃ f₁' (↑ x) f₂' (↑ y ⊞ f₂))
    ins x (♥node t₁ y t₂ s)     | inj₂ y≤x with ins x t₂
    ins x (♥node {f₁}t₁ y t₂ s) | inj₂ y≤x    | ♠case f₁' f₂' refl (c , t) = ♥case (f₁ ⊞ ↑ y ⊞ f₁') f₂' (sym (£⊞₂ f₁ (↑ y) f₁' f₂'))
                                                                             (♠node t₁ y t (£ins₂ f₁ f₁' f₂' s (sortedOf t) y≤x)
                                                                              ▷ sym (£⊞₂ f₁ (↑ y) f₁' (↑ x ⊞ f₂')))
    ins x (♠node t₁ y t₂ s)  with x ≤A? y
    ins x (♠node t₁ y t₂ s)     | inj₁ x≤y with ins x t₁
    ins x (♠node t₁ y {f₂}t₂ s) | inj₁ x≤y    | ♥case f₁' f₂' refl t       = ♠case f₁' (f₂' ⊞ ↑ y ⊞ f₂) (++assoc f₁' f₂' (↑ y ⊞ f₂))
                                                                             (bal₁ t y t₂ (£ins₁ f₁' f₂' f₂ s (sortedOf t) x≤y)
                                                                              Σ▷ £⊞₂ f₁' (↑ x) f₂' (↑ y ⊞ f₂))
    ins x (♠node t₁ y {f₂}t₂ s) | inj₁ x≤y    | ♠case f₁' f₂' refl (c , t) = ♠case f₁' (f₂' ⊞ ↑ y ⊞ f₂) (++assoc f₁' f₂' (↑ y ⊞ f₂))
                                                                             (, ♠node t y t₂ (£ins₁ f₁' f₂' f₂ s (sortedOf t) x≤y)
                                                                              ▷ £⊞₂ f₁' (↑ x) f₂' (↑ y ⊞ f₂))
    ins x (♠node t₁ y t₂ s)     | inj₂ y≤x with ins x t₂
    ins x (♠node {f₁}t₁ y t₂ s) | inj₂ y≤x    | ♥case f₁' f₂' refl t       = ♠case (f₁ ⊞ ↑ y ⊞ f₁') f₂' (sym (£⊞₂ f₁ (↑ y) f₁' f₂'))
                                                                             (bal₂ t₁ y t (£ins₂ f₁ f₁' f₂' s (sortedOf t) y≤x)
                                                                              Σ▷ sym (£⊞₂ f₁ (↑ y) f₁' (↑ x ⊞ f₂')))
    ins x (♠node {f₁}t₁ y t₂ s) | inj₂ y≤x    | ♠case f₁' f₂' refl (c , t) = ♠case (f₁ ⊞ ↑ y ⊞ f₁') f₂' (sym (£⊞₂ f₁ (↑ y) f₁' f₂'))
                                                                             (, ♠node t₁ y t (£ins₂ f₁ f₁' f₂' s (sortedOf t) y≤x)
                                                                              ▷ sym (£⊞₂ f₁ (↑ y) f₁' (↑ x ⊞ f₂')))

    abstract
      £sort₃ : (f₁ f₂ f₃ : F) → sorted (f₁ ⊞ f₂ ⊞ f₃)
                              → sorted (f₁ ⊞ f₂)
      £sort₃ f₁ f₂ f₃ s = £sort₁ (f₁ ⊞ f₂) f₃ (s s▷ sym (++assoc f₁ f₂ f₃))

    -- rewrite a ∉ flattening
    infixl 20 _▷∉_
    _▷∉_ : {k : K}{f₁ f₂ : F} → k ∉ f₁ → f₁ ≡ f₂ → k ∉ f₂
    z ▷∉ refl = z

    --  a comprehensive return type for fnd
    data FndR : (k : K)(f : F)(c : C)(n : ℕ) → Set where
      ∉case : {k : K}{f : F}(kf : k ∉ f) -- the "key not found in tree" case
              {c : C}{n : ℕ}                         → FndR k f c n
      ♥case :               (f₁' : F)
                            (x : A)
                            (f₂' : F)
              {k : K}       (kx : k ≡ † x)
              {f : F}       (e : f ≡ f₁' ⊞ ↑ x ⊞ f₂')
              {n : ℕ}                                → FndR k f ♥ n
      ♠case :               (f₁' : F)
                            (x : A)
                            (f₂' : F)
              {k : K}       (kx : k ≡ † x)
              {f : F}       (e : f ≡ f₁' ⊞ ↑ x ⊞ f₂')
              {n : ℕ}                                → FndR k f ♠ (suc n)

    -- Find element with key k
    fnd : (k : K){f : F}{c : C}{n : ℕ} → RB f c n → FndR k f c n
    fnd k leaf = ∉case (£∉[] k)
    fnd k (♥node t₁ d t₂ s)      with k <K? † d
    fnd k (♥node t₁ d t₂ s)         | tri< lt _ _ with fnd k t₁
    fnd k (♥node {f₁}t₁ d {f₂}t₂ s) | tri< lt _ _    | ∉case kf                = ∉case (£∉++ k kf (£∉< f₂ lt (£sort₂ f₁ (↑ d ⊞ f₂) s)))
    fnd k (♥node t₁ d {f₂}t₂ s)     | tri< lt _ _    | ♠case f₁' x f₂' kx refl = ♥case f₁' x (f₂' ⊞ ↑ d ⊞ f₂) kx (£⊞₂ f₁' (↑ x) f₂' (↑ d ⊞ f₂))
    fnd k (♥node {f₁}t₁ d {f₂}t₂ s) | tri≈ _ eq _                              = ♥case f₁ d f₂ eq refl
    fnd k (♥node t₁ d t₂ s)         | tri> _ _ gt with fnd k t₂
    fnd k (♥node {f₁}t₁ d {f₂}t₂ s) | tri> _ _ gt    | ∉case kf                = ∉case (£∉++ k (£∉> f₁ gt (£sort₃ f₁ (↑ d) f₂ s)) kf ▷∉ ++assoc f₁ (↑ d) f₂)
    fnd k (♥node {f₁}t₁ d t₂ s)     | tri> _ _ gt    | ♠case f₁' x f₂' kx refl = ♥case (f₁ ⊞ ↑ d ⊞ f₁') x f₂' kx (sym (£⊞₂ f₁ (↑ d) f₁' (↑ x ⊞ f₂')))
    fnd k (♠node t₁ d t₂ s)      with k <K? † d
    fnd k (♠node t₁ d t₂ s)         | tri< lt _ _ with fnd k t₁
    fnd k (♠node {f₁}t₁ d {f₂}t₂ s) | tri< lt _ _    | ∉case kf                = ∉case (£∉++ k kf (£∉< f₂ lt (£sort₂ f₁ (↑ d ⊞ f₂) s)))
    fnd k (♠node t₁ d {f₂}t₂ s)     | tri< lt _ _    | ♥case f₁' x f₂' kx refl = ♠case f₁' x (f₂' ⊞ ↑ d ⊞ f₂) kx (£⊞₂ f₁' (↑ x) f₂' (↑ d ⊞ f₂))
    fnd k (♠node t₁ d {f₂}t₂ s)     | tri< lt _ _    | ♠case f₁' x f₂' kx refl = ♠case f₁' x (f₂' ⊞ ↑ d ⊞ f₂) kx (£⊞₂ f₁' (↑ x) f₂' (↑ d ⊞ f₂))
    fnd k (♠node {f₁}t₁ d {f₂}t₂ s) | tri≈ _ eq _                              = ♠case f₁ d f₂ eq refl
    fnd k (♠node t₁ d t₂ s)         | tri> _ _ gt with fnd k t₂
    fnd k (♠node {f₁}t₁ d {f₂}t₂ s) | tri> _ _ gt    | ∉case kf                = ∉case (£∉++ k (£∉> f₁ gt (£sort₃ f₁ (↑ d) f₂ s)) kf ▷∉ ++assoc f₁ (↑ d) f₂)
    fnd k (♠node {f₁}t₁ d t₂ s)     | tri> _ _ gt    | ♥case f₁' x f₂' kx refl = ♠case (f₁ ⊞ ↑ d ⊞ f₁') x f₂' kx (sym (£⊞₂ f₁ (↑ d) f₁' (↑ x ⊞ f₂')))
    fnd k (♠node {f₁}t₁ d t₂ s)     | tri> _ _ gt    | ♠case f₁' x f₂' kx refl = ♠case (f₁ ⊞ ↑ d ⊞ f₁') x f₂' kx (sym (£⊞₂ f₁ (↑ d) f₁' (↑ x ⊞ f₂')))

    abstract
      £sort₄ : (f₁ f₂ f₃ f₄ : F) → sorted (f₁ ⊞ f₂ ⊞ f₃ ⊞ f₄) 
                                 → sorted (f₁ ⊞ f₂ ⊞ f₃)
      £sort₄ f₁ f₂ f₃ f₄ s = (£sort₁ ((f₁ ⊞ f₂) ⊞ f₃) f₄ (s s▷ sym (++assoc f₁ f₂ (f₃ ⊞ f₄)) s▷ sym (++assoc (f₁ ⊞ f₂) f₃ f₄))) s▷ ++assoc f₁ f₂ f₃

    -- additional re-balancing function used by delmin and del
    bal₃ : {f₀ : F}{n : ℕ} → RB f₀ ♠ n            →
                             (y : A)              →
           {f : F} →         RB f ♥ (suc n)       →
                             sorted (f₀ ⊞ ↑ y ⊞ f) → RB (f₀ ⊞ ↑ y ⊞ f) ♠ (suc (suc n))
    bal₃ {f₀}t₀ y (♥node {f₁}t₁ z {f₂}t₂ s') s
      with bal₂ t₀ y t₁ (£sort₄ f₀ (↑ y) f₁ (↑ z ⊞ f₂) s) 
    ... | c , B rewrite sym (£⊞₂ f₀ (↑ y) f₁ (↑ z ⊞ f₂)) = ♠node B z t₂ s

    -- convert any color node to a black node at either the same or +1 black level
    c⇛♠+ : {c : C}{f : F}{n : ℕ} → RB f c n → RB f ♠ n ⊎ RB f ♠ (suc n)
    c⇛♠+ {♥}r = inj₂ (♥⇛♠ r)
    c⇛♠+ {♠}r = inj₁ r

    --  a comprehensive return type for delmin
    data DelminR : (f : F)(c : C)(n : ℕ) → Set where
      lcase :                                             DelminR [] ♠ 0
      ♥case :        (m : A)
                     (f' : F)
              {f : F}(e : f ≡ ↑ m ⊞ f')
              {n : ℕ}(ct : Σ[ c ∶ C ](RB f' c n))       → DelminR f ♥ n
      ♠case :        (m : A)
                     (f' : F)
              {f : F}(e : f ≡ ↑ m ⊞ f')
              {n : ℕ}(ts : RB f' ♠ n ⊎ RB f' ♠ (suc n)) → DelminR f ♠ (suc n)

    -- Delete the minimum element, if there is one
    delmin : {n : ℕ}{c : C}{f : F} → RB f c n → DelminR f c n
    delmin leaf = lcase
    delmin (♥node t₁ d t₂ s)     with delmin t₁
    delmin (♥node t₁ d {f₂}t₂ s)    | lcase                                                        = ♥case d f₂ ([]++ (↑ d ⊞ f₂)) (, t₂)
    delmin (♥node t₁ d {f₂}t₂ s)    | ♠case m f' refl (inj₁ t) rewrite ++assoc (↑ m) f' (↑ d ⊞ f₂) = ♥case m (f' ⊞ ↑ d ⊞ f₂) refl 
                                                                                                      (bal₂ t d t₂              (£sort₂ (↑ m) (f' ⊞ ↑ d ⊞ f₂) s))
    delmin (♥node t₁ d {f₂}t₂ s)    | ♠case m f' refl (inj₂ t) rewrite ++assoc (↑ m) f' (↑ d ⊞ f₂) = ♥case m (f' ⊞ ↑ d ⊞ f₂) refl 
                                                                                                      (, ♥node t d t₂           (£sort₂ (↑ m) (f' ⊞ ↑ d ⊞ f₂) s))
    delmin (♠node t₁ d t₂ s)     with delmin t₁
    delmin (♠node t₁ d {f₂}t₂ s)    | lcase                                                        = ♠case d f₂ ([]++ (↑ d ⊞ f₂)) (c⇛♠+ t₂)
    delmin (♠node t₁ d {f₂}t₂ s)    | ♥case m f' refl (c , t)  rewrite ++assoc (↑ m) f' (↑ d ⊞ f₂) = ♠case m (f' ⊞ ↑ d ⊞ f₂) refl 
                                                                                                      (inj₂ (♠node t d t₂       (£sort₂ (↑ m) (f' ⊞ ↑ d ⊞ f₂) s)))
    delmin (♠node t₁ d {f₂}{♥}t₂ s) | ♠case m f' refl (inj₁ t) rewrite ++assoc (↑ m) f' (↑ d ⊞ f₂) = ♠case m (f' ⊞ ↑ d ⊞ f₂) refl (inj₂ 
                                                                                                      (bal₃ t d t₂              (£sort₂ (↑ m) (f' ⊞ ↑ d ⊞ f₂) s)))
    delmin (♠node t₁ d {f₂}{♠}t₂ s) | ♠case m f' refl (inj₁ t) rewrite ++assoc (↑ m) f' (↑ d ⊞ f₂) = ♠case m (f' ⊞ ↑ d ⊞ f₂) refl 
                                                                                                      (c⇛♠+ (proj₂ (bal₂ t d t₂ (£sort₂ (↑ m) (f' ⊞ ↑ d ⊞ f₂) s))))
    delmin (♠node t₁ d {f₂}t₂ s)    | ♠case m f' refl (inj₂ t) rewrite ++assoc (↑ m) f' (↑ d ⊞ f₂) = ♠case m (f' ⊞ ↑ d ⊞ f₂) refl 
                                                                                                      (inj₂ (♠node t d t₂       (£sort₂ (↑ m) (f' ⊞ ↑ d ⊞ f₂) s)))

    -- additional re-balancing function used by del
    bal₄ : {f : F}{n : ℕ} → RB f ♥ (suc n)       →
                            (y : A)              →
           {f₃ : F} →       RB f₃ ♠ n            →
                            sorted (f ⊞ ↑ y ⊞ f₃) → RB (f ⊞ ↑ y ⊞ f₃) ♠ (suc (suc n))
    bal₄ (♥node {f₁}t₁ x {f₂}t₂ s') y {f₃}t₃ s rewrite ++assoc f₁ (↑ x ⊞ f₂) (↑ y ⊞ f₃) | ++assoc (↑ x) f₂ (↑ y ⊞ f₃)
      with bal₁ t₂ y t₃ (£sort₂ (↑ x) (f₂ ⊞ ↑ y ⊞ f₃) (£sort₂ f₁ (↑ x ⊞ f₂ ⊞ ↑ y ⊞ f₃) s))
    ... | c , B = ♠node t₁ x B s

    --  a comprehensive return type for del (note similarity to FndR)
    data DelR : (k : K)(f : F)(c : C)(n : ℕ) → Set where
      ∉case : {k : K}{f : F}(kf : k ∉ f) -- the "key not found in tree" case
              {c : C}{n : ℕ}                                                     → DelR k f c n
      ♥case :               (f₁' : F)
                            (x : A)
                            (f₂' : F)
              {k : K}       (kx : k ≡ † x)
              {f : F}       (e : f ≡ f₁' ⊞ ↑ x ⊞ f₂')
              {n : ℕ}       (ct : Σ[ c ∶ C ](RB (f₁' ⊞ f₂') c n))                → DelR k f ♥ n
      ♠case :               (f₁' : F)
                            (x : A)
                            (f₂' : F)
              {k : K}       (kx : k ≡ † x)
              {f : F}       (e : f ≡ f₁' ⊞ ↑ x ⊞ f₂')
              {n : ℕ}       (tt : RB (f₁' ⊞ f₂') ♠ n ⊎ RB (f₁' ⊞ f₂') ♠ (suc n)) → DelR k f ♠ (suc n)

    -- Delete an element with key k
    del :                         (k : K)
          {f : F}{c : C}{n : ℕ} → RB f c n → DelR k f c n
    del k leaf = ∉case (£∉[] k)
    del k (♥node t₁ d t₂ s)      with k <K? † d
    del k (♥node t₁ d t₂ s)         | tri< lt _ _ with del k t₁
    del k (♥node {f₁}t₁ d {f₂}t₂ s) | tri< lt _ _    | ∉case kf                          = ∉case (£∉++ k kf (£∉< f₂ lt (£sort₂ f₁ (↑ d ⊞ f₂) s)))
    del k (♥node t₁ d {f₂}t₂ s)     | tri< lt _ _    | ♠case f₁' x f₂' kx refl (inj₁ t') = ♥case f₁' x (f₂' ⊞ ↑ d ⊞ f₂) kx (£⊞₂ f₁' (↑ x) f₂' (↑ d ⊞ f₂))
                                                                                           (bal₂ t' d t₂ (£del₁ f₁' (↑ x) f₂' (↑ d ⊞ f₂) s) Σ▷ ++assoc f₁' f₂' (↑ d ⊞ f₂))
    del k (♥node t₁ d {f₂}t₂ s)     | tri< lt _ _    | ♠case f₁' x f₂' kx refl (inj₂ t') = ♥case f₁' x (f₂' ⊞ ↑ d ⊞ f₂) kx (£⊞₂ f₁' (↑ x) f₂' (↑ d ⊞ f₂))
                                                                                           (, ♥node t' d t₂ (£del₁ f₁' (↑ x) f₂' (↑ d ⊞ f₂) s) ▷ ++assoc f₁' f₂' (↑ d ⊞ f₂))
    del k (♥node t₁ d t₂ s)         | tri≈ _ eq _ with delmin t₂
    del k (♥node {f₁}t₁ d t₂ s)     | tri≈ _ eq _    | lcase                             = ♥case f₁ d [] eq refl (, t₁ ▷ sym (++[] f₁))
    del k (♥node {f₁}t₁ d t₂ s)     | tri≈ _ eq _    | ♠case m f' refl (inj₁ t')         = ♥case f₁ d (↑ m ⊞ f') eq refl (bal₁ t₁ m t' (£del₂ f₁ (↑ d) (↑ m ⊞ f') s))
    del k (♥node {f₁}t₁ d t₂ s)     | tri≈ _ eq _    | ♠case m f' refl (inj₂ t')         = ♥case f₁ d (↑ m ⊞ f') eq refl (, ♥node t₁ m t' (£del₂ f₁ (↑ d) (↑ m ⊞ f') s))
    del k (♥node t₁ d t₂ s)         | tri> _ _ gt with del k t₂
    del k (♥node {f₁}t₁ d {f₂}t₂ s) | tri> _ _ gt    | ∉case kf rewrite sym (++assoc f₁ (↑ d) f₂) = ∉case (£∉++ k (£∉> f₁ gt (£sort₁ (f₁ ⊞ ↑ d) f₂ s)) kf)
    del k (♥node {f₁}t₁ d t₂ s)     | tri> _ _ gt    | ♠case f₁' x f₂' kx refl (inj₁ t') = ♥case (f₁ ⊞ ↑ d ⊞ f₁') x f₂' kx (sym (£⊞₂ f₁ (↑ d) f₁' (↑ x ⊞ f₂')))
                                                                                           (bal₁ t₁ d t' (£del₃ f₁ (↑ d) f₁' (↑ x) f₂' s) Σ▷ sym (£⊞₂ f₁ (↑ d) f₁' f₂'))
    del k (♥node {f₁}t₁ d t₂ s)     | tri> _ _ gt    | ♠case f₁' x f₂' kx refl (inj₂ t') = ♥case (f₁ ⊞ ↑ d ⊞ f₁') x f₂' kx (sym (£⊞₂ f₁ (↑ d) f₁' (↑ x ⊞ f₂')))
                                                                                           (, ♥node t₁ d t' (£del₃ f₁ (↑ d) f₁' (↑ x) f₂' s) ▷ sym (£⊞₂ f₁ (↑ d) f₁' f₂'))
    del k (♠node t₁ d t₂ s)      with k <K? † d
    del k (♠node t₁ d t₂ s)         | tri< lt _ _ with del k t₁
    del k (♠node {f₁}t₁ d {f₂}t₂ s) | tri< lt _ _    | ∉case kf                          = ∉case (£∉++ k kf (£∉< f₂ lt (£sort₂ f₁ (↑ d ⊞ f₂) s)))
    del k (♠node t₁ d {f₂}t₂ s)     | tri< lt _ _    | ♥case f₁' x f₂' kx refl (c' , t') = ♠case f₁' x (f₂' ⊞ ↑ d ⊞ f₂) kx (£⊞₂ f₁' (↑ x) f₂' (↑ d ⊞ f₂))
                                                                                           (inj₂ (♠node t' d t₂ (£del₁ f₁' (↑ x) f₂' (↑ d ⊞ f₂) s) ▷ ++assoc f₁' f₂' (↑ d ⊞ f₂)))
    del k (♠node t₁ d {f₂}{♥}t₂ s)  | tri< lt _ _    | ♠case f₁' x f₂' kx refl (inj₁ t') = ♠case f₁' x (f₂' ⊞ ↑ d ⊞ f₂) kx (£⊞₂ f₁' (↑ x) f₂' (↑ d ⊞ f₂))
                                                                                           (inj₂ (bal₃ t' d t₂ (£del₁ f₁' (↑ x) f₂' (↑ d ⊞ f₂) s) ▷ ++assoc f₁' f₂' (↑ d ⊞ f₂)))
    del k (♠node t₁ d {f₂}{♠}t₂ s)  | tri< lt _ _    | ♠case f₁' x f₂' kx refl (inj₁ t') = ♠case f₁' x (f₂' ⊞ ↑ d ⊞ f₂) kx (£⊞₂ f₁' (↑ x) f₂' (↑ d ⊞ f₂))
                                                                                           (c⇛♠+ (proj₂ (bal₂ t' d t₂ (£del₁ f₁' (↑ x) f₂' (↑ d ⊞ f₂) s)) ▷ ++assoc f₁' f₂' (↑ d ⊞ f₂)))
    del k (♠node t₁ d {f₂}t₂ s)     | tri< lt _ _    | ♠case f₁' x f₂' kx refl (inj₂ t') = ♠case f₁' x (f₂' ⊞ ↑ d ⊞ f₂) kx (£⊞₂ f₁' (↑ x) f₂' (↑ d ⊞ f₂))
                                                                                           (inj₂ (♠node t' d t₂ (£del₁ f₁' (↑ x) f₂' (↑ d ⊞ f₂) s) ▷ ++assoc f₁' f₂' (↑ d ⊞ f₂)))
    del k (♠node t₁ d t₂ s)         | tri≈ _ eq _ with delmin t₂
    del k (♠node {f₁}t₁ d t₂ s)     | tri≈ _ eq _    | lcase                             = ♠case f₁ d [] eq refl (c⇛♠+ (t₁ ▷ sym (++[] f₁)))
    del k (♠node {f₁}t₁ d t₂ s)     | tri≈ _ eq _    | ♥case m f' refl (c' , t')         = ♠case f₁ d (↑ m ⊞ f') eq refl
                                                                                           (inj₂ (♠node t₁ m t' (£del₂ f₁ (↑ d) (↑ m ⊞ f') s)))
    del k (♠node {f₁}{♥}t₁ d t₂ s)  | tri≈ _ eq _    | ♠case m f' refl (inj₁ t')         = ♠case f₁ d (↑ m ⊞ f') eq refl
                                                                                           (inj₂ (bal₄ t₁ m t' (£del₂ f₁ (↑ d) (↑ m ⊞ f') s)))
    del k (♠node {f₁}{♠}t₁ d t₂ s)  | tri≈ _ eq _    | ♠case m f' refl (inj₁ t')         = ♠case f₁ d (↑ m ⊞ f') eq refl
                                                                                           (c⇛♠+ (proj₂ (bal₁ t₁ m t' (£del₂ f₁ (↑ d) (↑ m ⊞ f') s))))
    del k (♠node {f₁}t₁ d t₂ s)     | tri≈ _ eq _    | ♠case m f' refl (inj₂ t')         = ♠case f₁ d (↑ m ⊞ f') eq refl
                                                                                           (inj₂ (♠node t₁ m t' (£del₂ f₁ (↑ d) (↑ m ⊞ f') s)))
    del k (♠node t₁ d t₂ s)         | tri> _ _ gt with del k t₂
    del k (♠node {f₁}t₁ d {f₂}t₂ s) | tri> _ _ gt    | ∉case kf rewrite sym (++assoc f₁ (↑ d) f₂) = ∉case (£∉++ k (£∉> f₁ gt (£sort₁ (f₁ ⊞ ↑ d) f₂ s)) kf)
    del k (♠node {f₁}t₁ d t₂ s)     | tri> _ _ gt    | ♥case f₁' x f₂' kx refl (c' , t') = ♠case (f₁ ⊞ ↑ d ⊞ f₁') x f₂' kx (sym (£⊞₂ f₁ (↑ d) f₁' (↑ x ⊞ f₂')))
                                                                                           (inj₂ (♠node t₁ d t' (£del₃ f₁ (↑ d) f₁' (↑ x) f₂' s) ▷ sym (£⊞₂ f₁ (↑ d) f₁' f₂')))
    del k (♠node {f₁}{♥}t₁ d t₂ s)  | tri> _ _ gt    | ♠case f₁' x f₂' kx refl (inj₁ t') = ♠case (f₁ ⊞ ↑ d ⊞ f₁') x f₂' kx (sym (£⊞₂ f₁ (↑ d) f₁' (↑ x ⊞ f₂')))
                                                                                           (inj₂ (bal₄ t₁ d t' (£del₃ f₁ (↑ d) f₁' (↑ x) f₂' s) ▷ sym (£⊞₂ f₁ (↑ d) f₁' f₂')))
    del k (♠node {f₁}{♠}t₁ d t₂ s)  | tri> _ _ gt    | ♠case f₁' x f₂' kx refl (inj₁ t') = ♠case (f₁ ⊞ ↑ d ⊞ f₁') x f₂' kx (sym (£⊞₂ f₁ (↑ d) f₁' (↑ x ⊞ f₂')))
                                                                                           (c⇛♠+ (proj₂ (bal₁ t₁ d t' (£del₃ f₁ (↑ d) f₁' (↑ x) f₂' s)) ▷ sym (£⊞₂ f₁ (↑ d) f₁' f₂')))
    del k (♠node {f₁}t₁ d t₂ s)     | tri> _ _ gt    | ♠case f₁' x f₂' kx refl (inj₂ t') = ♠case (f₁ ⊞ ↑ d ⊞ f₁') x f₂' kx (sym (£⊞₂ f₁ (↑ d) f₁' (↑ x ⊞ f₂')))
                                                                                           (inj₂ (♠node t₁ d t' (£del₃ f₁ (↑ d) f₁' (↑ x) f₂' s) ▷ sym (£⊞₂ f₁ (↑ d) f₁' f₂')))

    -- end of module RedBlackInternal

  private
    module Witness where
      -- Use real lists (and ≤A for sorted) to demonstrate that
      -- RedBlackInternal manipulates F and sorted terms consistent
      -- with their interpretation as lists - proving all of the
      -- lemma parameters:
      open import Data.List

      F : Set
      F = List A

      ↑_ : A → F
      ↑ a = a ∷ []

      []++ : (f : F) → [] ++ f ≡ f
      []++ f = refl

      ++[] : (f : F) → f ++ [] ≡ f
      ++[] [] = refl
      ++[] (x ∷ xs) rewrite ++[] xs = refl

      ++assoc : (a b c : F) → (a ++ b) ++ c ≡  a ++ (b ++ c)
      ++assoc [] b c = refl
      ++assoc (x ∷ xs) b c rewrite ++assoc xs b c = refl

      sorted : F → Set
      sorted [] = ⊤
      sorted (a ∷ []) = ⊤
      sorted (a ∷ b ∷ f) = a ≤A b × sorted (b ∷ f)

      sorted[] : sorted []
      sorted[] = tt

      sorted1 : (a : A) → sorted (↑ a)
      sorted1 a = tt

      £sort₁ : (f₁ f₂ : F) → sorted (f₁ ++ f₂) → sorted  f₁
      £sort₁ [] f₂ s = tt
      £sort₁ (x ∷ []) f₂ s = tt
      £sort₁ (x ∷ x' ∷ xs) f₂ (s , ss) = s , £sort₁ (x' ∷ xs) f₂ ss

      £sort₂ : (f₁ f₂ : F) → sorted (f₁ ++ f₂) → sorted f₂
      £sort₂ [] f₂ s = s
      £sort₂ (x ∷ []) [] s = tt
      £sort₂ (x ∷ []) (x' ∷ xs) (s , ss) = ss
      £sort₂ (x ∷ x' ∷ xs) f₂ (s , ss) = £sort₂ (x' ∷ xs) f₂ ss

      £ins₁ : (f₁ f₂ f₃ : F){x y : A} → sorted ((f₁ ++         f₂)  ++ (↑ y ++ f₃))
                                      → sorted  (f₁ ++ (↑ x ++ f₂))
                            → x ≤A y  → sorted ((f₁ ++ (↑ x ++ f₂)) ++ (↑ y ++ f₃))
      £ins₁ [] [] f₃ s₁ s₂ x≤y = x≤y , s₁
      £ins₁ [] (f₂ ∷ f₂s) f₃ s₁ (s₂ , s₂s) x≤y = s₂ , s₁
      £ins₁ (f₁ ∷ []) [] f₃ (s₁ , s₁s) (s₂ , s₂s) x≤y = s₂ , x≤y , s₁s
      £ins₁ (f₁ ∷ []) (f₂ ∷ f₂s) f₃ (s₁ , s₁s) (s₂ , s₂' , s₂s) x≤y = s₂ , s₂' , s₁s
      £ins₁ (f₁ ∷ f₁' ∷ f₁s) f₂ f₃ (s₁ , s₁s) (s₂ , s₂s) x≤y = s₂ , £ins₁ (f₁' ∷ f₁s) f₂ f₃ s₁s s₂s x≤y

      £ins₂ : (f₁ f₂ f₃ : F){y x : A} → sorted (f₁ ++ (↑ y ++ (f₂ ++         f₃)))
                                      → sorted                (f₂ ++ (↑ x ++ f₃))
                            → y ≤A x  → sorted (f₁ ++ (↑ y ++ (f₂ ++ (↑ x ++ f₃))))
      £ins₂ [] [] f₃ s₁ s₂ y≤x = y≤x , s₂
      £ins₂ [] (f₂ ∷ f₂s) f₃ (s₁ , s₁s) s₂ y≤x = s₁ , s₂
      £ins₂ (f₁ ∷ []) [] f₃ (s₁ , s₁s) s₂ y≤x = s₁ , y≤x , s₂
      £ins₂ (f₁ ∷ []) (f₂ ∷ f₂s) f₃ (s₁ , s₁' , s₁s) s₂ y≤x = s₁ , s₁' , s₂
      £ins₂ (f₁ ∷ f₁' ∷ f₁s) f₂ f₃ (s₁ , s₁s) s₂ y≤x = s₁ , £ins₂ (f₁' ∷ f₁s) f₂ f₃ s₁s s₂ y≤x

      £sortab : (a b : A)(f : F) → sorted (a ∷ f) → b ≤A a → sorted (b ∷ f)
      £sortab a b [] s ba = tt
      £sortab a b (x ∷ xs) (s₁ , s) ba = ≤Atrans ba s₁ , s

      £del₁ : (f₁ f₂ f₃ f₄ : F) → sorted ((f₁ ++ (f₂ ++ f₃)) ++ f₄)
                                → sorted ((f₁ ++        f₃)  ++ f₄)
      £del₁ [] [] f₃ f₄ s = s
      £del₁ [] (x ∷ []) [] f₄ s = £sort₂ (x ∷ []) f₄ s
      £del₁ [] (x ∷ []) (x' ∷ xs) f₄ (s , ss) = ss
      £del₁ [] (x ∷ x' ∷ xs) f₃ f₄ (s , ss) = £del₁ [] (x' ∷ xs) f₃ f₄ ss
      £del₁ (x ∷ []) [] f₃ f₄ s = s
      £del₁ (x ∷ []) (x' ∷ f₂) f₃ f₄ (s , ss) = £del₁ (↑ x) f₂ f₃ f₄ (£sortab x' x ((f₂ ++ f₃) ++ f₄) ss s)
      £del₁ (x ∷ x' ∷ xs) f₂ f₃ f₄ (s , ss) = s , £del₁ (x' ∷ xs) f₂ f₃ f₄ ss

      £del₂ : (f₁ f₂ f₃ : F) → sorted (f₁ ++ (f₂ ++ f₃))
                             → sorted (f₁ ++        f₃)
      £del₂ [] f₂ f₃ s = £sort₂ f₂ f₃ s
      £del₂ (x ∷ []) [] f₃ s = s
      £del₂ (x ∷ []) (x' ∷ xs) f₃ (s , ss) = £del₂ (↑ x) xs f₃ (£sortab x' x (xs ++ f₃) ss s)
      £del₂ (x ∷ x' ∷ xs) f₂ f₃ (s , ss) = s , £del₂ (x' ∷ xs) f₂ f₃ ss

      £del₃ : (f₁ f₂ f₃ f₄ f₅ : F) → sorted (f₁ ++ (f₂ ++ (f₃ ++ (f₄ ++ f₅))))
                                   → sorted (f₁ ++ (f₂ ++ (f₃ ++        f₅)))
      £del₃ [] [] f₃ f₄ f₅ s = £del₂ f₃ f₄ f₅ s
      £del₃ [] (x ∷ []) f₃ f₄ f₅ s = £del₂ (x ∷ f₃) f₄ f₅ s
      £del₃ [] (x ∷ x' ∷ xs) f₃ f₄ f₅ (s , ss) = s , £del₃ (↑ x') xs f₃ f₄ f₅ ss
      £del₃ (x ∷ []) [] f₃ f₄ f₅ s = £del₂ (x ∷ f₃) f₄ f₅ s
      £del₃ (x ∷ []) (x' ∷ xs) f₃ f₄ f₅ (s , ss) = s , £del₃ (↑ x') xs f₃ f₄ f₅ ss
      £del₃ (x ∷ x' ∷ xs) f₂ f₃ f₄ f₅ (s , ss) = s , £del₃ (x' ∷ xs) f₂ f₃ f₄ f₅ ss

      open import Data.Empty

      _∉_ : (k : K)(f : F) → Set
      k ∉ [] = ⊤
      k ∉ (x ∷ f) = ((k ≡ † x) → ⊥) × (k ∉ f)

      £∉[] : (k : K) → k ∉ []
      £∉[] k = tt

      imp< : {k₁ k₂ : K} → k₁ <K k₂ → k₁ ≡ k₂ → ⊥
      imp< {k₁}{k₂}le with k₁ <K? k₂
      imp< le | tri< a ¬b ¬c = ¬b
      imp< le | tri≈ ¬a b ¬c = λ _ → ¬a le
      imp< le | tri> ¬a ¬b c = ¬b

      flipimp : {k₁ k₂ : K} → (k₁ ≡ k₂ → ⊥) → k₂ ≡ k₁ → ⊥
      flipimp ne refl = ne refl

      £∉++ : (k : K){f₁ f₂ : F} → k ∉ f₁ → k ∉ f₂ → k ∉ (f₁ ++ f₂)
      £∉++ k {[]} ni₁ ni₂ = ni₂
      £∉++ k {x ∷ xs}{f₂} (s₁ , ss₁) ni₂ = s₁ ,(£∉++ k{xs}{f₂} ss₁ ni₂)

      £∉<₀ : {k : K}{x : A}{f : F} → k <K † x → sorted (↑ x ++ f) → k ∉ f
      £∉<₀ {k} {x} {[]} le s = tt
      £∉<₀ {k} {x} {x' ∷ xs} le (inj₁ x0 , ss) = (imp< (<Ktrans le x0)) , (£∉<₀ {k}{x'}{xs} (<Ktrans le x0) ss)
      £∉<₀ {k} {x} {x' ∷ xs} le (inj₂ y , ss) rewrite y = (imp< le) ,(£∉<₀ {k}{x'}{xs} le ss)

      £∉< : {k : K}{x : A}(f : F) → k <K † x → sorted (↑ x ++ f) → k ∉ (↑ x ++ f)
      £∉< {k}{x}f le s = (imp< le) , £∉<₀ {k}{x}{f} le s

      £sort< : {a b : A}{k : K}{f₁ f₂ : F} → sorted (a ∷ f₁ ++ b ∷ f₂) → † b <K k → † a <K k
      £sort< {a} {b} {k} {[]} (inj₁ x , ss) le = <Ktrans x le
      £sort< {a} {b} {k} {[]} (inj₂ y , ss) le rewrite y = le
      £sort< {a} {b} {k} {x ∷ xs} (inj₁ s , ss) le = <Ktrans s (£sort< {x}{b}{k}{xs}ss le)
      £sort< {a} {b} {k} {x ∷ xs} (inj₂ s , ss) le rewrite s = £sort< {x}{b}{k}{xs}ss le

      £∉>₀ : {k : K}{x : A}{f : F} → † x <K k → sorted (f ++ ↑ x) → k ∉ f
      £∉>₀ {k} {x} {[]} le s = tt
      £∉>₀ {k} {x} {a ∷ f} le s = flipimp (imp< (£sort< {a}{x}{k}{f}{[]} s le)) , £∉>₀ {k}{x}{f} le (£sort₂ (↑ a) (f ++ ↑ x) s)

      £∉> : {k : K}{x : A}(f : F) → † x <K k → sorted (f ++ ↑ x) → k ∉ (f ++ ↑ x)
      £∉> {k}{x}f le s = £∉++ k {f}{↑ x} (£∉>₀ {k}{x}{f}le s) (flipimp (imp< le) , tt)

      module RedBlackWitness = RedBlackInternal F ↑_ _++_ [] []++ ++[] ++assoc sorted sorted[] sorted1 £sort₁ £sort₂ £ins₁ £ins₂ _∉_ £∉[] £∉< £∉> £∉++ £del₁ £del₂ £del₃
      -- end of module Witness

  private
    -- use erased definitions (all → ⊤) for the public runtime Red-Black Tree
    F : Set
    F = ⊤

    ↑_ : A → F
    ↑ a = tt

    _++_ : F → F → F
    f₁ ++ f₂ = tt

    [] : F
    [] = tt

    []++ : (f : F) → [] ++ f ≡ f
    []++ f = refl

    ++[] : (f : F) → f ++ [] ≡ f
    ++[] f = refl

    ++assoc : (a b c : F) → (a ++ b) ++ c ≡ a ++ (b ++ c)
    ++assoc a b c = refl

    sorted : F → Set
    sorted f = ⊤

    sorted[] : sorted []
    sorted[] = tt

    sorted1 : (a : A) → sorted (↑ a)
    sorted1 a = tt

    £sort₁ : (f₁ f₂ : F) → sorted (f₁ ++ f₂) → sorted  f₁
    £sort₁ f₁ f₂ s = tt

    £sort₂ : (f₁ f₂ : F) → sorted (f₁ ++ f₂) → sorted f₂
    £sort₂ f₁ f₂ s = tt

    £ins₁ : (f₁ f₂ f₃ : F){x y : A} → sorted ((f₁ ++         f₂)  ++ (↑ y ++ f₃))
                                    → sorted  (f₁ ++ (↑ x ++ f₂))
                          → x ≤A y  → sorted ((f₁ ++ (↑ x ++ f₂)) ++ (↑ y ++ f₃))
    £ins₁ f₁ f₂ f₃ s₁ s₂ le = tt

    £ins₂ : (f₁ f₂ f₃ : F){y x : A} → sorted (f₁ ++ (↑ y ++ (f₂ ++         f₃)))
                                    → sorted                (f₂ ++ (↑ x ++ f₃))
                          → y ≤A x  → sorted (f₁ ++ (↑ y ++ (f₂ ++ (↑ x ++ f₃))))
    £ins₂ f₁ f₂ f₃ s₁ s₂ le = tt

    _∉_ : (k : K)(f : F) → Set
    k ∉ f = ⊤

    £∉[] : (k : K) → k ∉ []
    £∉[] k = tt

    £∉< : {k : K}{x : A}(f : F) → k <K † x → sorted (↑ x ++ f) → k ∉ (↑ x ++ f)
    £∉< f lt s = tt

    £∉> : {k : K}{x : A}(f : F) → † x <K k → sorted (f ++ ↑ x) → k ∉ (f ++ ↑ x)
    £∉> f lt s = tt

    £∉++ : (k : K){f₁ f₂ : F} → k ∉ f₁ → k ∉ f₂ → k ∉ (f₁ ++ f₂)
    £∉++ k ni₁ ni₂ = tt

    £del₁ : (f₁ f₂ f₃ f₄ : F) → sorted ((f₁ ++ (f₂ ++ f₃)) ++ f₄)
                              → sorted ((f₁ ++        f₃)  ++ f₄)
    £del₁ f₁ f₂ f₃ f₄ s = tt

    £del₂ : (f₁ f₂ f₃ : F) → sorted (f₁ ++ (f₂ ++ f₃))
                           → sorted (f₁ ++        f₃)
    £del₂ f₁ f₂ f₃ s = tt

    £del₃ : (f₁ f₂ f₃ f₄ f₅ : F) → sorted (f₁ ++ (f₂ ++ (f₃ ++ (f₄ ++ f₅))))
                                 → sorted (f₁ ++ (f₂ ++ (f₃ ++        f₅)))
    £del₃ f₁ f₂ f₃ f₄ f₅ s = tt

  module RedBlackErased = RedBlackInternal F ↑_ _++_ [] []++ ++[] ++assoc sorted sorted[] sorted1 £sort₁ £sort₂ £ins₁ £ins₂ _∉_ £∉[] £∉< £∉> £∉++ £del₁ £del₂ £del₃

  open RedBlackErased

  -- Public interface:
  Tree : Set
  Tree = Σ[ f ∶ F ] Σ[ n ∶ ℕ ] (RB f ♠ n) -- the root is always black

  insert : A → Tree → Tree
  insert x (f , n , t) with ins x t
  insert x (_ , n , t)    | ♠case f₁ f₂ refl (♥ , t') = f₁ ++ (↑ x ++ f₂) , suc n , ♥⇛♠ t'
  insert x (_ , n , t)    | ♠case f₁ f₂ refl (♠ , t') = f₁ ++ (↑ x ++ f₂) ,     n , t'

  open import Data.Maybe

  find : K → Tree → Maybe A
  find k (f , n , t)    with fnd k t
  find k (f ,      n ,  t) | ∉case _          = nothing
  find k (f , (suc n) , t) | ♠case _ x _ _ _  = just x

  delete : K → Tree → (Maybe A) × Tree
  delete k (f , n , t)    with del k t
  delete k (f ,      n  , t) | ∉case _                       = nothing , f          ,     n , t
  delete k (f , (suc n) , t) | ♠case f₁' x f₂' _ _ (inj₁ t') = just x  , f₁' ++ f₂' ,     n , t'
  delete k (f , (suc n) , t) | ♠case f₁' x f₂' _ _ (inj₂ t') = just x  , f₁' ++ f₂' , suc n , t'
